package com.streambase.contrib.currency.graph;

import java.util.List;

import org.jgrapht.graph.DefaultDirectedWeightedGraph;
import org.jgrapht.graph.DefaultWeightedEdge;


public class CurrencyGraph<V> {

	
// ************************************************************************ //
// This should only be set to true if we are testing this class	
	private static final boolean TEST = false;	
// ************************************************************************ //
	
	// Constants
	private static final double TRANSACTION_COSTS = 0.0;
	
	// Variables
	private DefaultDirectedWeightedGraph<String, DefaultWeightedEdge> g;
	private List<String> arbSequence;
	private double transactionCosts;	
	private boolean isArbitrage;	

/****************************************************************************************************************************************************/
/**************************************************************** CONSTRUCTORS **********************************************************************/
/****************************************************************************************************************************************************/
	
	/**
	 * /**
	 * The class constructor. 
	 * 	- It builds a new directed graph
	 *  - Adds all the relevant vertices (currencies)
	 *  - Adds the edges between currency pairs 
	 *  - Sets the original weight of each of the edges. 
	 *
	 * @param tCos the transaction cost percentage
	 * @param trackAll true to track all arb opportunities, false to track most profitable 
	 */
	public CurrencyGraph(double tCost){
		
		// Instantiate a new DirectedWeighted
    	this.g = new DefaultDirectedWeightedGraph<String, DefaultWeightedEdge>(DefaultWeightedEdge.class);
    	
    	// Initialize the graph depending on whether this a test or a real run
    	if (TEST){ setTestGraph();} else {setRealGraph();}
    	
    	// Initialize all the variables we will need later 
    	this.arbSequence = null;
    	this.isArbitrage = false;
    	
    	// Set the default transaction cost
    	this.transactionCosts = (tCost >= 0) ? tCost : TRANSACTION_COSTS; 
    	
	}
	
	
	/**
	 * /**
	 * The class constructor. 
	 * 	- It builds a new directed graph
	 *  - Adds all the relevant vertices (currencies)
	 *  - Adds the edges between currency pairs 
	 *  - Sets the original weight of each of the edges. 
	 *
	 * @param tCos the transaction cost percentage
	 * @param trackAll true to track all arb opportunities, false to track most profitable 
	 */
	public CurrencyGraph(){
		
		// Instantiate a new DirectedWeighted
    	this.g = new DefaultDirectedWeightedGraph<String, DefaultWeightedEdge>(DefaultWeightedEdge.class);
    	
    	// Initialize the graph depending on whether this a test or a real run
    	if (TEST){ setTestGraph();} else {setRealGraph();}
    	
    	// Initialize all the variables we will need later 
    	this.arbSequence = null;
    	this.isArbitrage = false;
    	
    	// Set the default transaction cost
    	this.transactionCosts = TRANSACTION_COSTS; 
    	
	}	

/****************************************************************************************************************************************************/
/*************************************************************** PUBLIC METHODS *********************************************************************/
/****************************************************************************************************************************************************/

	
	/**
	 * Set's the weight of an existing edge between two vertices.
	 * If the edge does not exist yet, it adds a new edge to the
	 * graph and sets the weight of the edge. 
	 * @param cur1 the source vertex
	 * @param cur2 the target vertex
	 * @param interest the conversion rate between two currencies
	 */
	public void setWeight(String cur1, String cur2, double interest){		
		if (this.g.containsEdge(cur1, cur2)){
			this.g.setEdgeWeight(this.g.getEdge(cur1, cur2), -Math.log(interest*(1.0 - this.transactionCosts)));
		} else {
			this.g.setEdgeWeight(this.g.addEdge(cur1, cur2), -Math.log(interest*(1.0 - this.transactionCosts)));
		}
	}
	
	
	/**
	 * Set's the transaction cost to a new value (as a percentage of the interest rate)
	 * @param cost the new transaction cost percentage
	 * @return true if the percentage is valid, false otherwise
	 */
	public boolean setTransactionCost(Double cost){
		if (cost == null || cost < 0.0 || cost > 1.0){			
			return false;
		}
		this.transactionCosts = cost;
		return true;	
	}
	
	
	/**
	 * Determine if there is Arbitrage in the exchange rates by detecting a cycle.
	 * Note that if there are no cycles, an empty set is returned. 
	 * @return the set of nodes involved in the cycle
	 */
	public void calculateArbitrage(){
		CycleDetector<String, DefaultWeightedEdge> b = new CycleDetector<String, DefaultWeightedEdge>(g,"ROOT");
    	b.calculateArbitrage();
    	this.isArbitrage = b.hasCycle();
    	this.arbSequence = b.getCycle();    
	}	

	/**
	 * Evaluates if the graph contains an arbitrage sequence  
     * @return true if the graph contains aribtage, false otherwise
	 */
	public boolean isArbitrage(){
		return this.isArbitrage;
	}

	/**
	 * Return the profit of the arbitrage sequence associated with 
	 * the cycle in the graph. Must call calculateArbitrage first. 
	 * @return a double
	 */
	public Double getProfit(){
		double cost = 1;
		if (this.isArbitrage){
			for (int i = 1; i < this.arbSequence.size(); i++){
				cost *= getWeight(this.arbSequence.get(i-1), this.arbSequence.get(i));
			}
		} 
		return cost;
	}
	
	/**
	 * Return the arbitrage sequence associated with the cycle in the graph. 
	 * Must call calculateArbitrage first. 
	 * @return a list of Strings
	 */
	public List<String> getArbSequence(){
		return this.arbSequence;
	}
	
		
/****************************************************************************************************************************************************/
/*************************************************************** PRIVATE METHODS ********************************************************************/
/****************************************************************************************************************************************************/
	
	/**
	 * Returns the weight of the edge
	 * @param cur1 the source vertex
	 * @param cur2 the target vertex
	 * @return the edge weight between the source and the target vertex
	 */
	private double getWeight(String cur1, String cur2){
		if (this.g.containsEdge(cur1, cur2)){
			return Math.exp(-this.g.getEdgeWeight(this.g.getEdge(cur1,cur2)))/(1.0 - this.transactionCosts);
		} else {
			return 0.0;
		}
	}
	
	
	
	/**
	 * Setup the test graph network
	 */
	private void setTestGraph(){
		// Add the vertices to the graph
        g.addVertex("USD");
        g.addVertex("EUR");
        g.addVertex("GBP");
        g.addVertex("JPY");
         
        // Add the root node
        g.addVertex("ROOT");
        
        // Add zero weight edges between the root and all other nodes   
        g.setEdgeWeight(g.addEdge("ROOT","USD"), 0);
        g.setEdgeWeight(g.addEdge("ROOT","EUR"), 0);
        g.setEdgeWeight(g.addEdge("ROOT","GBP"), 0);
        g.setEdgeWeight(g.addEdge("ROOT","JPY"), 0);		
	}
	
	
	/**
	 * Setup the real graph network
	 */
	private void setRealGraph(){		
		// Add the 24 vertices to the graph
        this.g.addVertex("AED");
        this.g.addVertex("AUD");
        this.g.addVertex("CAD");
        this.g.addVertex("CHF");
        this.g.addVertex("CZK");
        this.g.addVertex("DKK");
        this.g.addVertex("EUR");
        this.g.addVertex("GBP");
        this.g.addVertex("HKD");
        this.g.addVertex("HUF");
        this.g.addVertex("ILS");
        this.g.addVertex("JPY");
        this.g.addVertex("MXN");
        this.g.addVertex("NOK");
        this.g.addVertex("NZD");
        this.g.addVertex("PLN");
        this.g.addVertex("RON");
        this.g.addVertex("SAR");
        this.g.addVertex("SEK");
        this.g.addVertex("SGD");
        this.g.addVertex("THB");
        this.g.addVertex("TRY");
        this.g.addVertex("USD");
        this.g.addVertex("ZAR");
         
        // Add the root node
        this.g.addVertex("ROOT");
        
        // Add zero weight edges between the root and all 24 nodes   
        this.g.setEdgeWeight(this.g.addEdge("ROOT","AED"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","AUD"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","CAD"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","CHF"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","CZK"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","DKK"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","EUR"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","GBP"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","HKD"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","HUF"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","ILS"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","JPY"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","MXN"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","NOK"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","NZD"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","PLN"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","RON"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","SAR"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","SEK"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","SGD"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","THB"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","TRY"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","USD"), 0);
        this.g.setEdgeWeight(this.g.addEdge("ROOT","ZAR"), 0);
       			
	}
	
//	/**
//	 * Optional setup of the edges in the real graph network.
//	 * Note that there are a possible 120 edges 
//	 */
//	public void setRealEdges(){ 
//		this.g.setEdgeWeight(this.g.addEdge("AUD","CAD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","CHF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","NOK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","NZD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","THB"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","USD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("AUD","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","CHF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","NOK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CAD","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","NOK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","TRY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CHF","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CZK","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("CZK","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","THB"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("DKK","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","AUD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","CAD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","CHF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","GBP"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","HKD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","MXN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","NOK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","NZD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","RON"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","THB"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","TRY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","USD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("EUR","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","AUD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","CAD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","CHF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","NOK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","NZD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","THB"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","TRY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","USD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("GBP","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("HKD","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("MXN","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NOK","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NOK","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NOK","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","CAD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","CHF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","THB"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","USD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("NZD","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("PLN","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("PLN","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("PLN","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("SEK","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("SEK","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("SGD","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("THB","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("TRY","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","AED"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","CAD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","CHF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","CZK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","DKK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","HKD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","HUF"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","ILS"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","JPY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","MXN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","NOK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","PLN"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","RON"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","SAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","SEK"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","SGD"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","THB"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","TRY"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("USD","ZAR"), 0);
//		this.g.setEdgeWeight(this.g.addEdge("ZAR","JPY"), 0);
//	}
	
/****************************************************************************************************************************************************/
/****************************************************************** MAIN METHOD *********************************************************************/
/****************************************************************************************************************************************************/

		
	
	// For testing purposes
	public static void main(String [] args){
		List<String> arb;
		CurrencyGraph<String> cg = new CurrencyGraph<String>();
		cg.setWeight("USD","EUR",0.69546);
		cg.setWeight("USD","GBP",0.61482);
		cg.setWeight("USD","JPY",90.7400);
		cg.setWeight("EUR","USD",1.43790);
		cg.setWeight("EUR","GBP",0.88405);
		cg.setWeight("EUR","JPY",130.475);
		cg.setWeight("GBP","USD",1.62650);
		cg.setWeight("GBP","EUR",1.13116);
		cg.setWeight("GBP","JPY",147.589);
		cg.setWeight("JPY","USD",0.01102);
		cg.setWeight("JPY","EUR",0.00766);
		cg.setWeight("JPY","GBP",0.00678);
		cg.calculateArbitrage();
		arb = cg.getArbSequence();
		if (arb==null) {System.out.println(0);}
		else {										
			System.out.println("The arbitrage sequences " + arb);
			System.out.println("The profit is " + cg.getProfit());
		}


		
	}
			
}