package com.streambase.contrib.currency.graph;

import org.jgrapht.Graph;
import org.jgrapht.graph.DefaultDirectedWeightedGraph;
import org.jgrapht.graph.DefaultWeightedEdge;

import java.util.*;

/**
 * For reference, see http://en.wikipedia.org/wiki/Bellman%E2%80%93Ford_algorithm
 * @author Adam Diamant
 * @since August 2012
 */
public class CycleDetector<V, E> {

    private Graph<V, E> graph;
    private int sizeMinus1;
    private Map<V, Double> distance = null;
    private Map<V, V> predecessor;   
    private List<V> verticesInCycle;
    private List<V> cycle;

    /**
     * Constructor.
     *
     * @param graph A jGraphT representation of a graph 
     * @param source the root node
     */
    public CycleDetector(Graph<V, E> graph, V source) {
        this.graph = graph;
        this.distance = new HashMap<V, Double>();
        this.predecessor = new HashMap<V, V>();
        this.sizeMinus1 = graph.vertexSet().size() - 1;
        this.verticesInCycle = new LinkedList<V>();
        this.cycle = null;
        
        
        // Initialize the graph
        for (V v : graph.vertexSet()) {
        	this.distance.put(v, Double.POSITIVE_INFINITY);        	
        }
        this.distance.put(source, 0.0);            
        
        // Relax the edges repeatedly (|V| - 1 times)
        relaxEdges();
        
        // Relax the edges one more time to find negative weight cycles
        findCycles();
        
    }

    
    /**
     * The algorithm relaxes each edge in the graph |V| - 1 times,
     * where |V| is the number of vertices. After |V| - 1 relaxations 
     * on each edge, the shortest paths to each vertex will be found. 
     * During each iterations, both the shortest path and the predecessor
     * node is stored. When an edge is relaxed, the algorithm checks whether 
     * or not the shortest path estimate can be improved by going through a
     * different predecessor vertex. 
     */
    private void relaxEdges() {
    	V source, target; double weight;
    	for (int i = 0; i < this.sizeMinus1; i++){
    		for(E edge : this.graph.edgeSet()){
    			source = this.graph.getEdgeSource(edge);
    			target = this.graph.getEdgeTarget(edge);
    			weight = this.distance.get(source) + this.graph.getEdgeWeight(edge);
    			if (this.distance.get(target) > weight){
    				this.distance.put(target, weight);
    				this.predecessor.put(target, source);
    			}
    		}
    	}
    }
    
    /**
     * In a graph without negative-weight cycles, all shortest path values will be optimal 
     * after n - 1 iterations. In a graph with negative-weight cycles, the shortest 
     * path is undefined, since the cycle can be traversed an infinite number of 
     * times in order to decrease the path cost. Thus, to detect negative-weight cycles, 
     * we relax every edge one more time. If the shortest path estimate for one of the 
     * vertices improves, then there is a negative weight cycle.  
     * 
     */
    private void findCycles(){
    	V source, target; double weight;
    	for(E edge : this.graph.edgeSet()){
			source = this.graph.getEdgeSource(edge);
			target = this.graph.getEdgeTarget(edge);
			weight = this.distance.get(source) + this.graph.getEdgeWeight(edge);
			if (this.distance.get(target) > weight){
				this.distance.put(target, weight);
				this.predecessor.put(target, source);
				this.verticesInCycle.add(target);
			}
    	}
    }
    
    /**
     * Calculates the arbitrage sequence. The algorithm works by
     * iterating through each of the i target nodes (denoted by start(i)). 
     * By iterating through the predecessor graph, the nodes are traversed  
     * until the algorithm realizes that the current node has already been 
     * visited by the algorithm. If this node equals start(i) then the algo 
     * exists. Otherwise, we start at the i+1 target node.
     * 
     */
    public void calculateArbitrage(){
    	LinkedList<V> tempCycle; V currentNode;
    	for(V vertex : this.verticesInCycle){
    		tempCycle = new LinkedList<V>();
    		tempCycle.add(vertex);
    		currentNode = this.predecessor.get(vertex);    		
    		while(!tempCycle.contains(currentNode)){
    			tempCycle.addFirst(currentNode);
    			currentNode = this.predecessor.get(currentNode);
    		}
    		tempCycle.addFirst(currentNode);
    		if (currentNode.equals(vertex)){
    			this.cycle = tempCycle;    			
    			return;
    		}
    	}		
    }
    
    
    /**
     * Evaluates if the graph contains a negative cycle 
     * --------  Must call calculateArbitrage first! ------- 
     * @return true if the graph contains a negative cycle, false otherwise
     */
    public boolean hasCycle() {
        return this.cycle != null;
    }

    /**
     * Returns the edges of the negative cycle 
     * --------  Must call calculateArbitrage first! -------
     * @return a list of vertices in the cycle
     */
    public List<V> getCycle() {
        return this.cycle;
    }
    

    // This is just for testing purposes
    public static void main(String[] args) {
    	DefaultDirectedWeightedGraph<String, DefaultWeightedEdge> g;
    	g = new DefaultDirectedWeightedGraph<String, DefaultWeightedEdge>(DefaultWeightedEdge.class);
    	
    	String rootNode;
    	boolean testNumbers = false;
    	boolean testCycle = true;

    	if (testNumbers){
    		
    		// Set the root node
    		rootNode = "0";
    		
    		// Add the vertices
    		g.addVertex("0");
    		g.addVertex("1");
    		g.addVertex("2");
    		g.addVertex("3");
    		g.addVertex("4");
    		g.addVertex("5");
    		g.addVertex("6");
    		g.addVertex("7");  

    		if (testCycle){

    			// Add the edge weights
    			g.setEdgeWeight(g.addEdge("4","5"),0.35);
    			g.setEdgeWeight(g.addEdge("5","4"),-0.66);
    			g.setEdgeWeight(g.addEdge("4","7"),0.37);
    			g.setEdgeWeight(g.addEdge("5","7"),0.28);
    			g.setEdgeWeight(g.addEdge("7","5"),0.28);
    			g.setEdgeWeight(g.addEdge("5","1"),0.32);
    			g.setEdgeWeight(g.addEdge("0","4"),0.38);
    			g.setEdgeWeight(g.addEdge("0","2"),0.26);
    			g.setEdgeWeight(g.addEdge("7","3"),0.39);
    			g.setEdgeWeight(g.addEdge("1","3"),0.29);
    			g.setEdgeWeight(g.addEdge("2","7"),0.34);
    			g.setEdgeWeight(g.addEdge("6","2"),0.40);
    			g.setEdgeWeight(g.addEdge("3","6"),0.52);
    			g.setEdgeWeight(g.addEdge("6","0"),0.58);
    			g.setEdgeWeight(g.addEdge("6","4"),0.93);

    		} else {

    			// Add the edge weights
    			g.setEdgeWeight(g.addEdge("4","5"),0.35);
    			g.setEdgeWeight(g.addEdge("5","4"),0.35);
    			g.setEdgeWeight(g.addEdge("4","7"),0.37);
    			g.setEdgeWeight(g.addEdge("5","7"),0.28);
    			g.setEdgeWeight(g.addEdge("7","5"),0.28);
    			g.setEdgeWeight(g.addEdge("5","1"),0.32);
    			g.setEdgeWeight(g.addEdge("0","4"),0.38);
    			g.setEdgeWeight(g.addEdge("0","2"),0.26);
    			g.setEdgeWeight(g.addEdge("7","3"),0.39);
    			g.setEdgeWeight(g.addEdge("1","3"),0.29);
    			g.setEdgeWeight(g.addEdge("2","7"),0.34);
    			g.setEdgeWeight(g.addEdge("6","2"),-1.20);
    			g.setEdgeWeight(g.addEdge("3","6"),0.52);
    			g.setEdgeWeight(g.addEdge("6","0"),-1.40);
    			g.setEdgeWeight(g.addEdge("6","4"),-1.25);
    		}

    	} else {

    		// Set the root node
    		rootNode = "ROOT";
    		
    		// Add all the nodes
    		g.addVertex("ROOT");
    		g.addVertex("USD");
    		g.addVertex("EUR");
    		g.addVertex("GBP");
    		g.addVertex("JPY");

    		// Add zero weight edges between the root and all other nodes   
    		g.setEdgeWeight(g.addEdge("ROOT","USD"), 0);
    		g.setEdgeWeight(g.addEdge("ROOT","EUR"), 0);
    		g.setEdgeWeight(g.addEdge("ROOT","GBP"), 0);
    		g.setEdgeWeight(g.addEdge("ROOT","JPY"), 0);

    		// Add unit weight edges between the all valid currency pairs
    		g.setEdgeWeight(g.addEdge("USD","EUR"), -Math.log(0.69546));
    		g.setEdgeWeight(g.addEdge("USD","GBP"), -Math.log(0.61482));
    		g.setEdgeWeight(g.addEdge("USD","JPY"), -Math.log(90.7400));
    		g.setEdgeWeight(g.addEdge("EUR","USD"), -Math.log(1.43790));
    		g.setEdgeWeight(g.addEdge("EUR","GBP"), -Math.log(0.88405));
    		g.setEdgeWeight(g.addEdge("EUR","JPY"), -Math.log(130.475));
    		g.setEdgeWeight(g.addEdge("GBP","USD"), -Math.log(1.62650));
    		g.setEdgeWeight(g.addEdge("GBP","EUR"), -Math.log(1.13116));
    		g.setEdgeWeight(g.addEdge("GBP","JPY"), -Math.log(147.589));
    		g.setEdgeWeight(g.addEdge("JPY","USD"), -Math.log(0.01102));
    		g.setEdgeWeight(g.addEdge("JPY","EUR"), -Math.log(0.00766));
    		g.setEdgeWeight(g.addEdge("JPY","GBP"), -Math.log(0.00678));

    	}

    	CycleDetector<String, DefaultWeightedEdge> b = new CycleDetector<String, DefaultWeightedEdge>(g,rootNode);
    	b.calculateArbitrage();
    	boolean isCycle = b.hasCycle();
    	System.out.print("Does the graph have a cycle? ");
    	if (isCycle){
    		System.out.println("Yes");
    		System.out.println("The cycle is: " + b.getCycle());
    	} else {
    		System.out.println("No");
    	}
    }
}